#include <e.h>
#include "e_mod_win.h"
#include "e_slipshelf.h"
#include "e_slipwin.h"
#include "e_mod_layout.h"
#include "e_kbd.h"
#include "e_busywin.h"

/* FIXME: fix freerunner theme to have a slidesel theme that looks good */
/* FIXME: need some form of "busy" or "hourglass" display thing to pop up
 * while apps launch */

/* internal calls */
static void _cb_resize(void);
static void _cb_run(void *data);
static int _cb_event_border_add(void *data, int type, void *event);
static int _cb_event_border_remove(void *data, int type, void *event);
static int _cb_event_exe_del(void *data, int type, void *event);
static int _cb_run_timeout(void *data);
static int _have_borders(void);
static void _cb_slipshelf_home(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action);
static void _cb_slipshelf_close(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action);
static void _cb_slipshelf_apps(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action);
static void _cb_slipshelf_keyboard(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action);
static void _cb_slipwin_select(const void *data, E_Slipwin *esw, E_Border *bd);
static void _cb_slipshelf_select(const void *data, E_Slipshelf *ess, E_Border *bd);
static void _cb_slipshelf_add(const void *data, E_Slipshelf *ess, E_Border *bd);
static void _cb_slipshelf_del(const void *data, E_Slipshelf *ess, E_Border *bd);
static void _cb_slipshelf_home2(const void *data, E_Slipshelf *ess, E_Border *bd);
static int _cb_efreet_desktop_list_change(void *data, int type, void *event);
static int _cb_efreet_desktop_change(void *data, int type, void *event);
static void _apps_unpopulate(void);
static void _apps_populate(void);
static int _cb_update_deferred(void *data);

/* state */
static E_Zone *zone = NULL;
static Evas *evas = NULL;
static Evas_Object *sf = NULL;
static Evas_Object *bx = NULL;
static Evas_List *desks = NULL;
static Evas_List *handlers = NULL;
static Evas_List *sels = NULL;
static E_Slipshelf *slipshelf = NULL;
static E_Slipwin *slipwin = NULL;
static Ecore_Timer *defer = NULL;
static E_Kbd *vkbd = NULL;
static E_Busywin *busywin = NULL;

/* called from the module core */
void
_e_mod_win_init(E_Module *m)
{
   E_Container *con;

   zone = e_util_container_zone_number_get(0, 0);
   
   slipshelf = e_slipshelf_new(zone, e_module_dir_get(m));
   e_slipshelf_default_title_set(slipshelf, "ILLUME");
   if (!_have_borders())
     e_slipshelf_action_enabled_set(slipshelf, E_SLIPSHELF_ACTION_APPS, 0);
   e_slipshelf_action_callback_set(slipshelf, E_SLIPSHELF_ACTION_HOME,
				   _cb_slipshelf_home, NULL);
   e_slipshelf_action_callback_set(slipshelf, E_SLIPSHELF_ACTION_CLOSE,
				   _cb_slipshelf_close, NULL);
   e_slipshelf_action_callback_set(slipshelf, E_SLIPSHELF_ACTION_APPS,
				   _cb_slipshelf_apps, NULL);
   e_slipshelf_action_callback_set(slipshelf, E_SLIPSHELF_ACTION_KEYBOARD,
				   _cb_slipshelf_keyboard, NULL);
   e_slipshelf_border_select_callback_set(slipshelf, _cb_slipshelf_select, NULL);
   e_slipshelf_border_add_callback_set(slipshelf, _cb_slipshelf_add, NULL);
   e_slipshelf_border_del_callback_set(slipshelf, _cb_slipshelf_del, NULL);
   e_slipshelf_border_home_callback_set(slipshelf, _cb_slipshelf_home2, NULL);
   
   slipwin = e_slipwin_new(zone, e_module_dir_get(m));
   e_slipwin_border_select_callback_set(slipwin, _cb_slipwin_select, NULL);
   
   vkbd = e_kbd_new(zone, 
		    e_module_dir_get(m), 
		    e_module_dir_get(m),
		    e_module_dir_get(m));
   
   busywin = e_busywin_new(zone, e_module_dir_get(m));
   
   handlers = evas_list_append
     (handlers, ecore_event_handler_add
      (E_EVENT_BORDER_ADD, _cb_event_border_add, NULL));
   handlers = evas_list_append
     (handlers, ecore_event_handler_add
      (E_EVENT_BORDER_REMOVE, _cb_event_border_remove, NULL));
   handlers = evas_list_append
     (handlers, ecore_event_handler_add
      (ECORE_EXE_EVENT_DEL, _cb_event_exe_del, NULL));
   handlers = evas_list_append
     (handlers, ecore_event_handler_add
      (EFREET_EVENT_DESKTOP_LIST_CHANGE, _cb_efreet_desktop_list_change, NULL));
   handlers = evas_list_append
     (handlers, ecore_event_handler_add
      (EFREET_EVENT_DESKTOP_CHANGE, _cb_efreet_desktop_change, NULL));
   
   evas = zone->container->bg_evas;
   
   _apps_populate();


}

void
_e_mod_win_shutdown(void)
{
   _apps_unpopulate();
   evas_object_del(sf);
   evas_object_del(bx);
   e_object_del(E_OBJECT(slipshelf));
   e_object_del(E_OBJECT(slipwin));
   slipshelf = NULL;
   slipwin = NULL;
}

/* internal calls */
static void
_cb_resize(void)
{
   Evas_Coord mw, mh;
   int x, y, w, h;

   /* FIXME: this callback needs to be called whenever something changes with slipshelf config or zone size/config */
   e_slipshelf_safe_app_region_get(zone, &x, &y, &w, &h);
   e_box_min_size_get(bx, &mw, &mh);
   if (mw < w) mw = w;
   evas_object_move(sf, x, y);
   evas_object_resize(bx, mw, mh);
   evas_object_resize(sf, w, h);
}

typedef struct _Instance Instance;

struct _Instance
{
   Efreet_Desktop *desktop;
   E_Border       *border;
   Ecore_Timer    *timeout;
   int             startup_id;
   pid_t           pid;
   E_Busywin_Handle *handle;
};

static Evas_List *instances = NULL;

static void
_cb_run(void *data)
{
   Efreet_Desktop *desktop;
   Evas_List *l, *borders;
   E_Exec_Instance *eins;
   Instance *ins;
   char *exename, *p;
   
   desktop = data;
   printf("...RUN %s\n", desktop->exec);
   for (l = instances; l; l = l->next)
     {
	ins = l->data;
	if (ins->desktop == desktop)
	  {
	     if (ins->border)
	       {
		  printf("FLIPPED to EXE\n");
		  _e_mod_layout_border_show(ins->border);
	       }
	     return;
	  }
     }
   exename = NULL;
   p = strchr(desktop->exec, ' ');
   if (!p)
     {
	exename = strdup(desktop->exec);
     }
   else
     {
	exename = malloc(p - desktop->exec + 1);
	if (exename)
	  {
	     strncpy(exename, desktop->exec, p - desktop->exec);
	     exename[p - desktop->exec] = 0;
	  }
     }
   if (exename)
     {
	p = strrchr(exename, '/');
	if (p) strcpy(exename, p + 1);
     }
   printf("exe = %s\n", exename);
   borders = e_border_client_list();
   for (l = borders; l; l = l->next)
     {
	E_Border *bd;
	
	bd = l->data;
	printf("  ... has pid %i startup_id %i\n", bd->client.netwm.pid, bd->client.netwm.startup_id);
	if (e_exec_startup_id_pid_find(bd->client.netwm.pid,
				       bd->client.netwm.startup_id) == desktop)
	  {
	     printf("FLIPPED to EXE (backup1)\n");
	     _e_mod_layout_border_show(ins->border);
	     if (exename) free(exename);
	     return;
	  }
	if (bd->client.icccm.command.argv)
	  printf("  ... has argv0 = %s\n", bd->client.icccm.command.argv[0]);
	if ((exename) && (bd->client.icccm.command.argv) &&
	    ((!strcmp(exename, bd->client.icccm.command.argv[0]))))
	  {
	     printf("FLIPPED to EXE (backup2)\n");
	     _e_mod_layout_border_show(ins->border);
	     if (exename) free(exename);
	     return;
	  }
     }
   if (exename) free(exename);

   ins = calloc(1, sizeof(Instance));
   if (!ins) return;
   eins = e_exec(zone, desktop, NULL, NULL, "illume-launcher");
   printf("eins = %p\n", eins);
   ins->desktop = desktop;
   if (eins)
     {
	ins->startup_id = eins->startup_id;
	ins->pid = ecore_exe_pid_get(eins->exe);
     }
   printf("id = %i, pid = %i\n", ins->startup_id, ins->pid);
   ins->timeout = ecore_timer_add(20.0, _cb_run_timeout, ins);
   
     {
	char buf[256];
	
	snprintf(buf, sizeof(buf), "Starting %s", desktop->name);
	ins->handle = e_busywin_push(busywin, buf, NULL);
     }
   instances = evas_list_append(instances, ins);
}

static int 
_cb_event_border_add(void *data, int type, void *event)
{
   E_Event_Border_Add *ev;
   Efreet_Desktop *desktop;
   Instance *ins;
   Evas_List *l;
   
   ev = event;
   e_slipshelf_action_enabled_set(slipshelf, E_SLIPSHELF_ACTION_APPS, 1);
   desktop = e_exec_startup_id_pid_find(ev->border->client.netwm.pid,
					ev->border->client.netwm.startup_id);
   printf("ADD bd %p, desktop = %p, id = %i\n", ev->border, desktop,
	  ev->border->client.netwm.startup_id);
   for (l = instances; l; l = l->next)
     {
	ins = l->data;
	if (!ins->border)
	  {
	     if ((ins->startup_id == ev->border->client.netwm.startup_id) ||
		 (ins->pid == ev->border->client.netwm.pid)) 
	       {
		  ins->border = ev->border;
		  if (ins->handle)
		    {
		       e_busywin_pop(busywin, ins->handle);
		       ins->handle = NULL;
		    }
		  printf("found bd for ins\n");
		  if (ins->timeout) ecore_timer_del(ins->timeout);
		  ins->timeout = NULL;
		  // FIXME; if gadget is disabled, enable it
		  return 1;
	       }
	  }
     }
   return 1;
}

static int 
_cb_event_border_remove(void *data, int type, void *event)
{
   E_Event_Border_Remove *ev;
   Instance *ins;
   Evas_List *l;
   
   ev = event;
   printf("bd del %p, list = %p\n", ev->border, e_border_client_list());
   if (!_have_borders())
     e_slipshelf_action_enabled_set(slipshelf, E_SLIPSHELF_ACTION_APPS, 0);
   for (l = instances; l; l = l->next)
     {
	ins = l->data;
	if (ins->border == ev->border)
	  {
	     if (ins->handle)
	       {
		  e_busywin_pop(busywin, ins->handle);
		  ins->handle = NULL;
	       }
	     ins->border = NULL;
	     return 1;
	  }
     }
   return 1;
}

static int
_cb_event_exe_del(void *data, int type, void *event)
{
   Ecore_Exe_Event_Del *ev;
   Instance *ins;
   Evas_List *l;
   
   ev = event;
   for (l = instances; l; l = l->next)
     {
	ins = l->data;
	if (ins->pid == ev->pid)
	  {
	     if (ins->handle)
	       {
		  e_busywin_pop(busywin, ins->handle);
		  ins->handle = NULL;
	       }
             printf("remove pid\n");
	     instances = evas_list_remove_list(instances, l);
	     if (ins->timeout) ecore_timer_del(ins->timeout);
	     free(ins);
	     return 1;
	  }
     }
   return 1;
}

static int
_cb_run_timeout(void *data)
{
   Instance *ins;
   
   ins = data;
   if (ins->handle)
     {
	e_busywin_pop(busywin, ins->handle);
	ins->handle = NULL;
     }
   if (!ins->border)
     {
	instances = evas_list_remove(instances, ins);
	free(ins);
	return 0;
     }
   ins->timeout = NULL;
   return 0;
}

static int
_have_borders(void)
{
   Evas_List *borders, *l;
   int num = 0;
   
   borders = e_border_client_list();
   for (l = borders; l; l = l->next)
     {
	E_Border *bd;
	
	bd = l->data;
	if (e_object_is_del(E_OBJECT(bd))) continue;
	if ((!bd->client.icccm.accepts_focus) &&
	    (!bd->client.icccm.take_focus)) continue;
	if (bd->client.netwm.state.skip_taskbar) continue;
	if (bd->user_skip_winlist) continue;
	num++;
     }
   return num;
}

static void
_cb_slipshelf_home(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action)
{
   Evas_List *l, *borders;
   
   borders = e_border_client_list();
   for (l = borders; l; l = l->next)
     {
	E_Border *bd;
	
	bd = l->data;
	if (e_object_is_del(E_OBJECT(bd))) continue;
	if ((!bd->client.icccm.accepts_focus) &&
	    (!bd->client.icccm.take_focus)) continue;
	if (bd->client.netwm.state.skip_taskbar) continue;
	if (bd->user_skip_winlist) continue;
	_e_mod_layout_border_hide(bd);
     }
}

static void
_cb_slipshelf_close(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action)
{
   E_Border *bd;
   
   bd = e_border_focused_get();
   if (bd)
     {
	if (e_object_is_del(E_OBJECT(bd))) return;
	if ((!bd->client.icccm.accepts_focus) &&
	    (!bd->client.icccm.take_focus)) return;
	if (bd->client.netwm.state.skip_taskbar) return;
	if (bd->user_skip_winlist) return;
	_e_mod_layout_border_close(bd);
     }
}

static void
_cb_slipshelf_apps(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action)
{
   if (!_have_borders()) return;
   e_slipwin_show(slipwin);
}

static void
_cb_slipshelf_keyboard(const void *data, E_Slipshelf *ess, E_Slipshelf_Action action)
{
   printf("KBD toggle %i\n", (int)vkbd->out);
   if (vkbd->out)
     e_kbd_hide(vkbd);
   else
     e_kbd_show(vkbd);
}

static void
_cb_slipwin_select(const void *data, E_Slipwin *esw, E_Border *bd)
{
   if (bd) _e_mod_layout_border_show(bd);
}

static void
_cb_slipshelf_select(const void *data, E_Slipshelf *ess, E_Border *bd)
{
   if (bd) _e_mod_layout_border_show(bd);
}

static void
_cb_slipshelf_add(const void *data, E_Slipshelf *ess, E_Border *bd)
{
   printf("FIXME: implement add\n");
}

static void
_cb_slipshelf_del(const void *data, E_Slipshelf *ess, E_Border *bd)
{
   if (!bd) return;
   if (e_object_is_del(E_OBJECT(bd))) return;
   if ((!bd->client.icccm.accepts_focus) &&
       (!bd->client.icccm.take_focus)) return;
   if (bd->client.netwm.state.skip_taskbar) return;
   if (bd->user_skip_winlist) return;
   _e_mod_layout_border_close(bd);
}

static void
_cb_slipshelf_home2(const void *data, E_Slipshelf *ess, E_Border *bd)
{
   Evas_List *l, *borders;
   
   borders = e_border_client_list();
   for (l = borders; l; l = l->next)
     {
	E_Border *bd;
	
	bd = l->data;
	if (e_object_is_del(E_OBJECT(bd))) continue;
	if ((!bd->client.icccm.accepts_focus) &&
	    (!bd->client.icccm.take_focus)) continue;
	if (bd->client.netwm.state.skip_taskbar) continue;
	if (bd->user_skip_winlist) continue;
	_e_mod_layout_border_hide(bd);
     }
}

static void
_apps_unpopulate(void)
{
   while (sels)
     {
//	printf("DEL %p\n", sels->data);
	evas_object_del(sels->data);
	sels = evas_list_remove_list(sels, sels);
     }
   while (desks)
     {
	Efreet_Desktop *desktop;
	
	desktop = desks->data;
	efreet_desktop_free(desktop);
	desks = evas_list_remove_list(desks, desks);
     }
   if (bx) evas_object_del(bx);
   bx = NULL;
   if (sf) evas_object_del(sf);
   sf = NULL;
}

static void
_apps_populate(void)
{
   Evas_Coord mw, mh;
   Evas_Object *o;
   
   sf = e_scrollframe_add(evas);
   e_scrollframe_single_dir_set(sf, 1);
   evas_object_move(sf, zone->x, zone->y);
   evas_object_resize(sf, zone->w, zone->h);
   evas_object_show(sf);

   bx = e_box_add(evas);
   e_box_orientation_set(bx, 0);
   e_box_homogenous_set(bx, 1);

   e_box_freeze(bx);

     {
	Efreet_Menu *menu, *entry, *subentry;
	Efreet_Desktop *desktop;
	char *label, *icon, *plabel;
	char buf[1024];

	menu = efreet_menu_get();
	if (menu)
	  {
	     ecore_list_first_goto(menu->entries);
	     while ((entry = ecore_list_next(menu->entries)))
	       {
		  if (entry->type != EFREET_MENU_ENTRY_MENU) continue;
		  
		  plabel = NULL;
		  
		  desktop = entry->desktop;
		  if (entry->name) plabel = strdup(entry->name);
		  if (!plabel) plabel = strdup("???");
		  
		  o = e_slidesel_add(evas);
		  e_slidesel_item_distance_set(o, 128);
		  
		  ecore_list_first_goto(entry->entries);
		  while ((subentry = ecore_list_next(entry->entries)))
		    {
		       if (subentry->type != EFREET_MENU_ENTRY_DESKTOP) continue;

		       label = icon = NULL;
		       desktop = subentry->desktop;
		       if (desktop)
			 {
			    icon = ecore_hash_get(desktop->x, "X-Application-Screenshot");
			    if (icon) icon = strdup(icon);
			 }
		       if ((!icon) && (subentry->icon))
			 {
			    if (subentry->icon[0] == '/')
			      icon = strdup(subentry->icon);
			    else
			      icon = efreet_icon_path_find(e_config->icon_theme,
							   subentry->icon, "512x512");
			 }
		       if (subentry->name) label = strdup(subentry->name);
		       if (desktop)
			 {
			    if (!label)
			      label = strdup(desktop->generic_name);
			    
			    if ((!icon) && (desktop->icon))
			      icon = efreet_icon_path_find(e_config->icon_theme,
							   desktop->icon, "512x5152");
			 }
		  
		       if (!icon) icon = efreet_icon_path_find(e_config->icon_theme,
							  "hires.jpg", "512x512");
		       if (!icon) icon = strdup("DEFAULT");
		       if (!label) label = strdup("???");
		       
//		       printf("app: %s / %s\n icon: %s\n", plabel, label, icon);

		       snprintf(buf, sizeof(buf), "%s / %s", plabel, label);
		       desks = evas_list_append(desks, desktop);
		       efreet_desktop_ref(desktop);
		       e_slidesel_item_add(o, buf, icon, _cb_run, desktop);
		       if (label) free(label);
		       if (icon) free(icon);
		    }
		  if (plabel) free(plabel);
		  e_box_pack_end(bx, o);
		  e_box_pack_options_set(o, 1, 1, 0, 0, 0.5, 0.5, 
					 420, 215, 420, 215);
		  evas_object_show(o);
//		  printf("ADD %p\n", o);
		  sels = evas_list_append(sels, o);
	       }
	  }
     }
   
   e_box_thaw(bx);
   
   _cb_resize();
   
   e_scrollframe_child_set(sf, bx);
   evas_object_show(bx);
}

static int
_cb_efreet_desktop_list_change(void *data, int type, void *event)
{
   printf("list... change\n");
   if (defer) ecore_timer_del(defer);
   defer = ecore_timer_add(1.0, _cb_update_deferred, NULL);
   return 1;
}

static int
_cb_efreet_desktop_change(void *data, int type, void *event)
{
   printf("desktop... change\n");
   if (bx)
     {
	if (defer) ecore_timer_del(defer);
	defer = ecore_timer_add(1.0, _cb_update_deferred, NULL);
     }
   return 1;
}

static int
_cb_update_deferred(void *data)
{
   _apps_unpopulate();
   _apps_populate();
   printf("deferred update done\n");
   defer = NULL;
   return 0;
}

/* FIXME: have launcher slide each item using box offset left/right on 
 * launcher on/off */
